package uk.ac.bbsrc.babraham.SparkSpotter;
/*
 * SparkSpotter - An ImageJ plugin for finding bright sparks in movies
 * Copyright (c) 2007 Simon Andrews

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JOptionPane;
import javax.swing.UIManager;

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;

public class SparkSpotter_ implements PlugInFilter, Runnable {
	
	public static final String VERSION = "1.1";
	
	private ImagePlus imagePlus;
	private int [][][] pixels;
	private Vector<SparkEventGroup> eventGroups = new Vector<SparkEventGroup>();
	private Vector<SparkEvent> events = new Vector<SparkEvent>();
	private SparkSpotterPreferences prefs;
	private SparkSpotterProgressDialog progressBar;
	private boolean cancelCalculation = false;

	public SparkSpotter_ () {
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {}
	}
	
	public void run(ImageProcessor ip) {
	}
	
	public int setup(String arg, ImagePlus imagePlus) {
		
		if (imagePlus == null) {
			JOptionPane.showMessageDialog(null, "No image appears to be loaded.", "Can't run plugin", JOptionPane.INFORMATION_MESSAGE);
			return DONE;
		}
		
		this.imagePlus = imagePlus;
		if (arg.equals("about")) {
		    showAbout();
		    return DONE;
		}

		// The stack represents the full tiff stack of images
		ImageStack stack = imagePlus.getStack();
		
//		System.err.println("Found stack with "+stack.getSize()+" images of size "+stack.getHeight()+"x"+stack.getWidth());
		
		// We need to make up a data structure to hold all of the image
		// data for the whole stack so we can interate through it later.
		pixels = new int [stack.getHeight()][stack.getWidth()][stack.getSize()];
		
		// We now need to step through the images filling in the pixel array
		// Although this is an Object array it's actually an array of pixel arrays
		Object [] images = stack.getImageArray();
		
		for (int i=0;i<stack.getSize();i++) {
//			System.err.println("Processing image "+i);
			short [] thesePixels = new short[0];
			
			// Depending on what kind of image we've got we could have byte, short or int arrays
			if (images[i] instanceof short[]) {
				thesePixels = (short[])images[i];
			}
			else if (images[i] instanceof byte[]) {
				// Just map these into a short array
				byte [] b = (byte[])images[i];
				thesePixels = new short[b.length];
				for (int j=0;j<b.length;j++) {
					thesePixels[j] = (short)b[j];
				}
			}
			else if (images[i] instanceof int []) {
				JOptionPane.showMessageDialog(null, "This plugin needs 8 or 16 bit greyscale images", "Can't run plugin", JOptionPane.INFORMATION_MESSAGE);
				return DONE;
			}

			else {
				JOptionPane.showMessageDialog(null, "Didn't recognise the format of your images", "Can't run plugin", JOptionPane.ERROR_MESSAGE);
				return DONE;
			}

			
//			System.err.println("Found "+thesePixels.length+" pixels");
			
			// Fill in the appropriate part of the pixel array
			for (int y=0;y<pixels.length;y++) {
				for (int x=0;x<pixels[y].length;x++) {
//					System.err.println("Adding position "+x+","+y+" from position "+((y*pixels[y].length)+x));
					pixels[y][x][i] = (int)thesePixels[(y*pixels[y].length)+x];
				}
			}
		}
		
		prefs = new SparkSpotterPreferences(this);
		
		return DONE;
	}

	public SparkSpotterPreferences getPrefs() {
		return prefs;
	}
	
	public void cancelCalculation() {
		cancelCalculation = true;
		prefs.analysisFinished();
	}
	
	public void startAnalysis() {
		progressBar = new SparkSpotterProgressDialog(this);
		Thread t = new Thread(this);
		t.start();
	}
	
	private void findSparks() {
		
		// Clean up after the last run
		eventGroups.clear();
		events.clear();
		
		
		// We need to process the pixels one by one looking for interesting patterns
		for (int y=0;y<pixels.length;y++) {
			progressBar.updateProgress(y, pixels.length);
			if (cancelCalculation) {
				cancelCalculation = false;
				return;
			}
//			System.err.println("Looking for events in row "+y);
			PIXEL: for (int x=0;x<pixels[y].length;x++) {
//				System.err.println("Looking for events in pixel "+x+","+y);
				int [] data = new int [imagePlus.getStack().getSize()];
				int largest=0;
				int smallest=0;
				for (int s=0;s<data.length;s++) {
					
					// We might be looking at a range of pixels if they've extended the
					// search area.  We also need to be able to bail out if we can't
					// pull out the whole search area.
					int total=0;
					int count=0;
					for (int ya = y-prefs.searchArea();ya<=y+prefs.searchArea();ya++) {

						if (ya<0 || ya >= imagePlus.getHeight()) {
							// We can't get the whole analysed region
//							System.err.println("Skipping Y pixel "+ya+" from original "+y+" and area "+prefs.searchArea());
							continue PIXEL;
						}
						for (int xa = x-prefs.searchArea();xa<=x+prefs.searchArea();xa++) {

							if (xa<0 || xa >= imagePlus.getWidth()) {
//								System.err.println("Skipping X pixel "+xa+" from original "+x+" and area "+prefs.searchArea());
								// We can't get the whole analysed region
								continue PIXEL;
							}

							total += pixels[ya][xa][s];
							count++;
						}
					}
					
					data[s] = total/count;

					if (s==0) {
						largest = data[s];
						smallest = data[s];
					}
					else {
						if (data[s]>largest) largest = data[s];
						if (data[s]<smallest) smallest = data[s];
					}

				}
				
				// We can skip pixels which have little or no contrast
//				if (largest-smallest < 150) continue;
				
				findEvents(data,x,y);
			}
		}
		
		mergeEvents();
		progressBar.processingFinished();
		prefs.analysisFinished();
		new SparkDisplay((SparkEventGroup [])eventGroups.toArray(new SparkEventGroup [0]),imagePlus,prefs);

	}
	
	private void mergeEvents () {
		// Go through all of the events trying to assign them to a group
		
		Enumeration<SparkEvent> e = events.elements();
		EVENTS: while (e.hasMoreElements()) {
			SparkEvent se = e.nextElement();
			Enumeration<SparkEventGroup> eg = eventGroups.elements();
			while (eg.hasMoreElements()) {
				SparkEventGroup seg = eg.nextElement();
				if (seg.overlaps(se, prefs.searchArea(), 1)) {
					seg.addEvent(se);
					continue EVENTS;
				}
			}
			// If we get here there are no overlaps with existing groups
			SparkEventGroup seg = new SparkEventGroup(se);
			eventGroups.add(seg);
		}
		
		// Now we need to go through the groups trying to merge them together.
		while (true) {
			Vector<SparkEventGroup> newEventGroups = new Vector<SparkEventGroup>();
			Enumeration<SparkEventGroup> e2 = eventGroups.elements();
			OUTER: while (e2.hasMoreElements()) {
				SparkEventGroup seg = e2.nextElement();
				Enumeration<SparkEventGroup> e3 = newEventGroups.elements();
				while (e3.hasMoreElements()) {
					SparkEventGroup seg2 = e3.nextElement();
					if (seg.overlaps(seg2, prefs.searchArea(), 1)) {
						SparkEvent [] sparks = seg2.getEvents();
						for (int s=0;s<sparks.length;s++) {
							seg2.addEvent(sparks[s]);
							// Move on straight away
							continue OUTER;
						}
					}
				}
				newEventGroups.add(seg);
			}
			
//			System.err.println("Old events list="+eventGroups.size()+" New list="+newEventGroups.size());
			
			if (newEventGroups.size() == eventGroups.size()) break; // Nothing changed
			eventGroups = newEventGroups;
		}
		
	}
	
	private void findEvents (int [] data, int x, int y) {
		
		// Just for testing
//		if(x != 121 || y != 170) return;
		
		// First we need to generate a smoothed line through the
		// pixel data
		int [] corrected = RunningNoramlisation.normalise(data,prefs.windowSize());
		
		// Now we need to look for points which deviate significantly from this distribution
		
		// Significant events are those which are >4 standard deviations above the mean.
		double [] meanStdev = calculateStdev(corrected);
		
		SparkEvent currentEvent = null;
		
		for (int f=4;f<corrected.length-4;f++) {		
			
			// Calculate how many Stdevs above the mean we are
			double fold = (corrected[f]-meanStdev[0])/meanStdev[1];
					
			if (fold > prefs.foldCutoff() && corrected[f]>prefs.diffCutoff()) {
				if (currentEvent == null) {
					// Start a new event
					currentEvent = new SparkEvent(f,corrected[f],fold,x,y,corrected,data);
				}
				else {
					// Extend the current event
					currentEvent.continueEvent(f, corrected[f],fold);
				}
			}
			else {
				if (currentEvent != null) {
					events.add(currentEvent);
					currentEvent = null;
				}
			}
			
		}
		
	}
	
	private double [] calculateStdev (int [] data) {
		
		double mean=0;
		for (int i=0;i<data.length;i++) {
			mean += data[i];
		}
		mean /= data.length;
		
		double stdev = 0;
		for (int i=0;i<data.length;i++) {
			stdev += Math.pow(data[i]-mean,2);
		}
		stdev /= data.length;
		stdev = Math.sqrt(stdev);
		
		return new double [] {mean,stdev};

	}
	
	
	public void showAbout() {
		IJ.showMessage("About SparkSpotter_...",
				"Need to write something here..."
				);

	}


	public static void main (String [] args) {
		new SparkSpotter_();
	}

	public void run() {
		findSparks();
	}

	
}
